from importlib.metadata import PackageNotFoundError, version
from importlib.resources import files

__package_name__ = "meds_testing_helpers"
try:
    __version__ = version(__package_name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

GEN_YAML = files(__package_name__).joinpath("configs/generate_dataset.yaml")
INF_YAML = files(__package_name__).joinpath("configs/infer_dataset_config.yaml")

__all__ = [
    "GEN_YAML",
    "INF_YAML",
    "__version__",
    "__package_name__",
]



import subprocess
import tempfile
from pathlib import Path


def test_dataset_config_inference(simple_static_MEDS):
    """Test the inference of a config to generate a synthetic version of a dataset."""

    with tempfile.TemporaryDirectory() as temp_dir:
        output_fp = Path(temp_dir) / "generated_config.yaml"
        cmd = [
            "infer_MEDS_sample_gen_config",
            f"dataset_dir={str(simple_static_MEDS.resolve())}",
            f"output_fp={str(output_fp.resolve())}",
        ]

        out = subprocess.run(cmd, shell=False, check=False, capture_output=True)

        error_str = (
            f"Command failed with return code {out.returncode}.\n"
            f"Command stdout:\n{out.stdout.decode()}\n"
            f"Command stderr:\n{out.stderr.decode()}"
        )

        assert out.returncode == 0, error_str

        assert output_fp.exists(), f"Output file {output_fp} does not exist."


def test_dataset_config_inference_errors(simple_static_MEDS):
    """Test the inference of a config to generate a synthetic version of a dataset."""

    with tempfile.TemporaryDirectory() as temp_dir:
        output_fp = Path(temp_dir) / "generated_config.yaml"
        output_fp.touch()

        cmd = [
            "infer_MEDS_sample_gen_config",
            f"dataset_dir={str(simple_static_MEDS.resolve())}",
            f"output_fp={str(output_fp.resolve())}",
        ]

        out = subprocess.run(cmd, shell=False, check=False, capture_output=True)

        assert out.returncode != 0, "Should fail because output file exists."

        cmd.append("do_overwrite=True")
        out = subprocess.run(cmd, shell=False, check=False, capture_output=True)
        assert out.returncode == 0, "Should succeed because output file exists but overwrite is True."



"""This file ensures that the necessary modules are reloaded for accurate coverage tracking."""

import importlib

import meds_testing_helpers.pytest_plugin
import meds_testing_helpers.static_sample_data

importlib.reload(meds_testing_helpers.pytest_plugin)
importlib.reload(meds_testing_helpers.static_sample_data)



import json
import logging
from io import StringIO
from pathlib import Path
from typing import Any

from yaml import load as load_yaml

try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader

import polars as pl
import pyarrow as pa
import pyarrow.parquet as pq
from meds import (
    DatasetMetadata,
    code_field,
    code_metadata_filepath,
    code_metadata_schema,
    data_schema,
    data_subdirectory,
    dataset_metadata_filepath,
    description_field,
    label_schema,
    numeric_value_field,
    parent_codes_field,
    prediction_time_field,
    subject_id_field,
    subject_split_schema,
    subject_splits_filepath,
    time_field,
)

logger = logging.getLogger(__name__)


SHARDED_DF_T = dict[str, pl.DataFrame]


class MEDSDataset:
    """A minimal helper class for working with MEDS Datasets intended for use in testing, not production.

    This class is intended to be used for testing and development purposes only. It is not intended to be used
    in production code, and as such has only been optimized for small datasets, and may not support all
    aspects of the MEDS schema.

    Attributes:
        root_dir: The root directory of the dataset, if provided. If specified, data will be read from this
            root directory upon access, not stored in memory. If not provided, the below parameters must be
            provided upon initialization.
        data_shards: A dictionary of data shards, where the keys are the shard names and the values are the
            data tables. Upon access of this attribute, the data will be returned as pyarrow tables. Upon
            specification in the constructor, polars dataframes are expected instead.
        dataset_metadata: The metadata for the dataset, stored as a DatasetMetadata object (which is just a
            type-annotated dictionary).
        code_metadata: The metadata for the codes. Upon access of this attribute, the data will be returned as
            a pyarrow table. Upon specification in the constructor, a polars dataframe is expected instead.
        subject_splits: The subject splits for the dataset. Optional. Upon access of this attribute, the data
            will be returned as a pyarrow table. Upon specification in the constructor, a polars dataframe is
            expected instead. If not specified for an otherwise valid dataset, `None` will be returned.
        task_labels: Optionally, you can also include task labels. These are not formally connected to a
            particular MEDS dataset (in terms of storage on disk; see
            https://github.com/Medical-Event-Data-Standard/meds/issues/75 for more information), you can also
            track a collection of sharded task labels by task name in this class. If specified, this should be
            a dictionary mapping task name to sharded task files, in the proper MEDS label dataframe format.

    Examples:
        >>> data_shards = {
        ...     "0": pl.DataFrame({"subject_id": [0], "time": [0], "numeric_value": [None], "code": ["A"]}),
        ...     "1": pl.DataFrame({"subject_id": [1], "time": [0], "numeric_value": [1.0], "code": ["B"]}),
        ... }
        >>> dataset_metadata = DatasetMetadata(
        ...     dataset_name="test",
        ...     dataset_version="0.0.1",
        ...     etl_name="foo",
        ...     etl_version="0.0.1",
        ...     meds_version="0.3.3",
        ...     created_at="1/1/2025",
        ...     extension_columns=[],
        ... )
        >>> code_metadata = pl.DataFrame({
        ...     "code": ["A", "B"], "description": ["foo", "bar"],
        ...     "parent_codes": pl.Series([None, None], dtype=pl.List(pl.Utf8)),
        ... })
        >>> subject_splits = None
        >>> task_labels = None
        >>> D = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits,
        ...     task_labels=task_labels,
        ... )
        >>> D # doctest: +NORMALIZE_WHITESPACE
        MEDSDataset(data_shards={'0': {'subject_id': [0],
                                       'time': [0],
                                       'numeric_value': [None],
                                       'code': ['A']},
                                 '1': {'subject_id': [1],
                                       'time': [0],
                                       'numeric_value': [1.0],
                                       'code': ['B']}},
                    dataset_metadata={'dataset_name': 'test',
                                      'dataset_version': '0.0.1',
                                      'etl_name': 'foo',
                                      'etl_version': '0.0.1',
                                      'meds_version': '0.3.3',
                                      'created_at': '1/1/2025',
                                      'extension_columns': []},
                    code_metadata={'code': ['A', 'B'],
                                   'description': ['foo', 'bar'],
                                   'parent_codes': [None, None]})
        >>> print(D)
        MEDSDataset:
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: [["A","B"]]
          description: [["foo","bar"]]
          parent_codes: [[null,null]]
        subject_splits: None
        >>> print(D.shard_fps)
        None

        Note that code metadata can be inferred to be empty if not provided:

        >>> print(MEDSDataset(data_shards=data_shards, dataset_metadata=dataset_metadata))
        MEDSDataset:
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: []
          description: []
          parent_codes: []
        subject_splits: None

        Note that when no task labels are provided, their properties are `None`:

        >>> print(D.task_names)
        None
        >>> print(D.task_labels)
        None

        There are also a collection of filepath related properties that are `None` when the root directory is
        not set:

        >>> print(D.task_root_dir)
        None
        >>> print(D.task_names_fp)
        None
        >>> print(D.task_label_fps)
        None
        >>> print(D.dataset_metadata_fp)
        None
        >>> print(D.code_metadata_fp)
        None
        >>> print(D.subject_splits_fp)
        None
        >>> print(D.shard_fps)
        None

        You can save and load datasets from disk in the proper format. Note that equality persists after this
        operation. Note that, when existent, filepath variables are then set as well. But, those parameters
        that depend on files existing when they don't may still be None

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     D2 = D.write(Path(tmpdir))
        ...     assert D == D2
        ...     print(f"repr: {repr(D2).replace(tmpdir, '...')}")
        ...     print(f"str: {str(D2).replace(tmpdir, '...')}")
        ...     print("|")
        ...     print("Filepaths:")
        ...     print(f"  task_root_dir: {str(D2.task_root_dir).replace(tmpdir, '...')}")
        ...     print(f"  task_names_fp: {str(D2.task_names_fp).replace(tmpdir, '...')}")
        ...     print(f"  dataset_metadata_fp: {str(D2.dataset_metadata_fp).replace(tmpdir, '...')}")
        ...     print(f"  code_metadata_fp: {str(D2.code_metadata_fp).replace(tmpdir, '...')}")
        ...     print(f"  subject_splits_fp: {str(D2.subject_splits_fp).replace(tmpdir, '...')}")
        ...     print(f"  shard_fps: {str(D2.shard_fps).replace(tmpdir, '...')}")
        ...     print(f"  task_label_fps: {D2.task_label_fps}")
        repr: MEDSDataset(root_dir=PosixPath('...'))
        str: MEDSDataset:
        stored in root_dir: ...
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: [["A","B"]]
          description: [["foo","bar"]]
          parent_codes: [[null,null]]
        subject_splits: None
        |
        Filepaths:
          task_root_dir: .../task_labels
          task_names_fp: .../task_labels/.task_names.json
          dataset_metadata_fp: .../metadata/dataset.json
          code_metadata_fp: .../metadata/codes.parquet
          subject_splits_fp: .../metadata/subject_splits.parquet
          shard_fps: [PosixPath('.../data/0.parquet'), PosixPath('.../data/1.parquet')]
          task_label_fps: None

        You can also add subject splits to the dataset:

        >>> subject_splits = pl.DataFrame({"subject_id": [0, 1], "split": ["train", "held_out"]})
        >>> D = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits,
        ... )
        >>> print(D)
        MEDSDataset:
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: [["A","B"]]
          description: [["foo","bar"]]
          parent_codes: [[null,null]]
        subject_splits:
          pyarrow.Table
          subject_id: int64
          split: string
          ----
          subject_id: [[0,1]]
          split: [["train","held_out"]]

        Here's an example with task labels. Note that, when provided, you can also query task names
        with a dedicated property (otherwise it is null). Task labels can be empty, or have empty shards. Note
        that root-dir related task properties can still be None even if tasks are set.

        >>> task_df_empty = pl.DataFrame(
        ...     {
        ...         "subject_id": [],
        ...         "prediction_time": [],
        ...         "boolean_value": [],
        ...         "integer_value": [],
        ...         "float_value": [],
        ...         "categorical_value": [],
        ...     },
        ...     schema=MEDSDataset.PL_LABEL_SCHEMA
        ... )
        >>> task_df_nonempty = pl.DataFrame(
        ...     {
        ...         "subject_id": [0, 1],
        ...         "prediction_time": [0, 10],
        ...         "boolean_value": [True, False],
        ...         "integer_value": [None, None],
        ...         "float_value": [None, None],
        ...         "categorical_value": [None, None],
        ...     },
        ...     schema=MEDSDataset.PL_LABEL_SCHEMA
        ... )
        >>> task_labels = { # Format is task_name -> shard -> dataframe
        ...     "A": {},
        ...     "B": {"0": task_df_empty},
        ...     "C": {"0": task_df_nonempty},
        ... }
        >>> D = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     task_labels=task_labels,
        ... )
        >>> D # doctest: +NORMALIZE_WHITESPACE
        MEDSDataset(data_shards={'0': {'subject_id': [0],
                                       'time': [0],
                                       'numeric_value': [None],
                                       'code': ['A']},
                                 '1': {'subject_id': [1],
                                       'time': [0],
                                       'numeric_value': [1.0],
                                       'code': ['B']}},
                    dataset_metadata={'dataset_name': 'test',
                                      'dataset_version': '0.0.1',
                                      'etl_name': 'foo',
                                      'etl_version': '0.0.1',
                                      'meds_version': '0.3.3',
                                      'created_at': '1/1/2025',
                                      'extension_columns': []},
                    code_metadata={'code': ['A', 'B'],
                                   'description': ['foo', 'bar'],
                                   'parent_codes': [None, None]},
                    task_labels={'A': {},
                                 'B': {'0': {'subject_id': [],
                                             'prediction_time': [],
                                             'boolean_value': [],
                                             'integer_value': [],
                                             'float_value': [],
                                             'categorical_value': []}},
                                 'C': {'0': {'subject_id': [0, 1],
                                             'prediction_time': [datetime.datetime(1970, 1, 1, 0, 0),
                                                                 datetime.datetime(1970, 1, 1, 0, 0, 0, 10)],
                                             'boolean_value': [True, False],
                                             'integer_value': [None, None],
                                             'float_value': [None, None],
                                             'categorical_value': [None, None]}}})
        >>> print(D)
        MEDSDataset:
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: [["A","B"]]
          description: [["foo","bar"]]
          parent_codes: [[null,null]]
        subject_splits: None
        task labels:
          * A:
          * B:
            - 0:
              pyarrow.Table
              subject_id: int64
              prediction_time: timestamp[us]
              boolean_value: bool
              integer_value: int64
              float_value: double
              categorical_value: string
              ----
              subject_id: [[]]
              prediction_time: [[]]
              boolean_value: [[]]
              integer_value: [[]]
              float_value: [[]]
              categorical_value: []
          * C:
            - 0:
              pyarrow.Table
              subject_id: int64
              prediction_time: timestamp[us]
              boolean_value: bool
              integer_value: int64
              float_value: double
              categorical_value: string
              ----
              subject_id: [[0,1]]
              prediction_time: [[1970-01-01 00:00:00.000000,1970-01-01 00:00:00.000010]]
              boolean_value: [[true,false]]
              integer_value: [[null,null]]
              float_value: [[null,null]]
              categorical_value: [[null,null]]
        >>> D.task_names
        ['A', 'B', 'C']

        Note that as we don't have a root dir, the file path parameters are still `None`

        >>> print(D.task_root_dir)
        None
        >>> print(D.task_label_fps)
        None

        Reading/Writing with task labels works identically as without:

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     D2 = D.write(Path(tmpdir))
        ...     assert D == D2
        ...     print(f"repr: {repr(D2).replace(tmpdir, '...')}")
        ...     print(f"str: {str(D2).replace(tmpdir, '...')}")
        repr: MEDSDataset(root_dir=PosixPath('...'))
        str: MEDSDataset:
        stored in root_dir: ...
        dataset_metadata:
          - dataset_name: test
          - dataset_version: 0.0.1
          - etl_name: foo
          - etl_version: 0.0.1
          - meds_version: 0.3.3
          - created_at: 1/1/2025
          - extension_columns: []
        data_shards:
          - 0:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[0]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["A"]]
            numeric_value: [[null]]
          - 1:
            pyarrow.Table
            subject_id: int64
            time: timestamp[us]
            code: string
            numeric_value: float
            ----
            subject_id: [[1]]
            time: [[1970-01-01 00:00:00.000000]]
            code: [["B"]]
            numeric_value: [[1]]
        code_metadata:
          pyarrow.Table
          code: string
          description: string
          parent_codes: list<item: string>
            child 0, item: string
          ----
          code: [["A","B"]]
          description: [["foo","bar"]]
          parent_codes: [[null,null]]
        subject_splits: None
        task labels:
          * A:
          * B:
            - 0:
              pyarrow.Table
              subject_id: int64
              prediction_time: timestamp[us]
              boolean_value: bool
              integer_value: int64
              float_value: double
              categorical_value: string
              ----
              subject_id: [[]]
              prediction_time: [[]]
              boolean_value: [[]]
              integer_value: [[]]
              float_value: [[]]
              categorical_value: []
          * C:
            - 0:
              pyarrow.Table
              subject_id: int64
              prediction_time: timestamp[us]
              boolean_value: bool
              integer_value: int64
              float_value: double
              categorical_value: string
              ----
              subject_id: [[0,1]]
              prediction_time: [[1970-01-01 00:00:00.000000,1970-01-01 00:00:00.000010]]
              boolean_value: [[true,false]]
              integer_value: [[null,null]]
              float_value: [[null,null]]
              categorical_value: [[null,null]]

        Equality is determined by the equality of the data, metadata, code metadata, subject splits, and task
        labels:

        >>> D1 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits
        ... )
        >>> D1 == "foobar"
        False
        >>> D2 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits
        ... )
        >>> D1 == D2
        True
        >>> D2 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=None,
        ... )
        >>> D1 == D2
        False
        >>> alt_data_shards = {
        ...     "0": pl.DataFrame({"subject_id": [1], "time": [0], "numeric_value": [None], "code": ["A"]}),
        ...     "1": pl.DataFrame({"subject_id": [1], "time": [0], "numeric_value": [1.0], "code": ["B"]}),
        ... }
        >>> D2 = MEDSDataset(
        ...     data_shards=alt_data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits
        ... )
        >>> D1 == D2
        False
        >>> alt_dataset_metadata = DatasetMetadata(
        ...     dataset_name="test_2",
        ...     dataset_version="0.0.1",
        ...     etl_name="foo",
        ...     etl_version="0.0.1",
        ...     meds_version="0.3.3",
        ...     created_at="1/1/2025",
        ...     extension_columns=[],
        ... )
        >>> D2 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=alt_dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits
        ... )
        >>> D1 == D2
        False
        >>> alt_code_metadata = pl.DataFrame({
        ...     "code": ["A", "B"], "description": ["bar", "foo"],
        ...     "parent_codes": pl.Series([None, None], dtype=pl.List(pl.Utf8)),
        ... })
        >>> D2 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=alt_code_metadata,
        ...     subject_splits=subject_splits
        ... )
        >>> D1 == D2
        False
        >>> D2 = MEDSDataset(
        ...     data_shards=data_shards,
        ...     dataset_metadata=dataset_metadata,
        ...     code_metadata=code_metadata,
        ...     subject_splits=subject_splits,
        ...     task_labels=task_labels,
        ... )
        >>> D1 == D2
        False

        Errors are raised in a number of circumstances:

        >>> MEDSDataset()
        Traceback (most recent call last):
            ...
        ValueError: data_shards must be provided if root_dir is None
        >>> MEDSDataset(data_shards=data_shards)
        Traceback (most recent call last):
            ...
        ValueError: dataset_metadata must be provided if root_dir is None
    """

    CSV_TS_FORMAT = "%m/%d/%Y, %H:%M:%S"

    PL_DATA_SCHEMA = {
        subject_id_field: pl.Int64,
        time_field: pl.Datetime("us"),
        code_field: pl.String,
        numeric_value_field: pl.Float32,
    }

    PL_CODE_METADATA_SCHEMA = {
        code_field: pl.String,
        description_field: pl.String,
        parent_codes_field: pl.List(pl.String),
    }

    PL_SUBJECT_SPLIT_SCHEMA = {
        subject_id_field: pl.Int64,
        "split": pl.String,
    }

    PL_LABEL_SCHEMA = {
        subject_id_field: pl.Int64,
        prediction_time_field: pl.Datetime("us"),
        "boolean_value": pl.Boolean,
        "integer_value": pl.Int64,
        "float_value": pl.Float64,
        "categorical_value": pl.String,
    }

    TIME_FIELDS = {time_field, prediction_time_field}

    TASK_LABELS_SUBDIR = "task_labels"
    TASK_NAMES_FN = ".task_names.json"

    def __init__(
        self,
        root_dir: Path | None = None,
        data_shards: SHARDED_DF_T | None = None,
        dataset_metadata: DatasetMetadata | None = None,
        code_metadata: pl.DataFrame | None = None,
        subject_splits: pl.DataFrame | None = None,
        task_labels: dict[str, SHARDED_DF_T] | None = None,
    ):
        if root_dir is None:
            if data_shards is None:
                raise ValueError("data_shards must be provided if root_dir is None")
            if dataset_metadata is None:
                raise ValueError("dataset_metadata must be provided if root_dir is None")
            if code_metadata is None:
                logger.warning("Inferring empty code metadata as none was provided.")
                code_metadata = pl.DataFrame(
                    {code_field: [], description_field: [], parent_codes_field: []},
                    schema=self.PL_CODE_METADATA_SCHEMA,
                )

        self.root_dir = root_dir
        self.data_shards = data_shards
        self.dataset_metadata = dataset_metadata
        self.code_metadata = code_metadata
        self.subject_splits = subject_splits
        self.task_labels = task_labels

        # These will throw errors if the data is malformed.
        self.data_shards
        self.code_metadata
        self.subject_splits
        self.dataset_metadata

    @classmethod
    def parse_csv(cls, csv: str, **schema_updates) -> pl.DataFrame:
        """Parses a CSV string into a MEDS-related dataframe using the provided schema.

        Args:
            csv: The CSV string to parse.
            schema_updates: The schema to use when parsing the CSV, passed as keyword arguments. Note that
                timestamp columns will be read as strings then converted to the requested type using the
                default CSV timestamp format. Schema defaults are drawn from all the available MEDS schemas,
                _without consideration for whether the passed csv is appropriate for that given schema_.

        Returns:
            A polars DataFrame with the parsed data.

        Raises:
            ValueError: If the CSV cannot be read under the provided schema and/or the schema defaults.

        Examples:
            >>> MEDSDataset.parse_csv(
            ...     'subject_id,time,code,numeric_value\\n0,"1/1/2025, 12:00:00",foo,1.0',
            ...     subject_id=pl.Int64, time=pl.Datetime("us"), code=pl.String, numeric_value=pl.Float32
            ... )
            shape: (1, 4)
            ┌────────────┬─────────────────────┬──────┬───────────────┐
            │ subject_id ┆ time                ┆ code ┆ numeric_value │
            │ ---        ┆ ---                 ┆ ---  ┆ ---           │
            │ i64        ┆ datetime[μs]        ┆ str  ┆ f32           │
            ╞════════════╪═════════════════════╪══════╪═══════════════╡
            │ 0          ┆ 2025-01-01 12:00:00 ┆ foo  ┆ 1.0           │
            └────────────┴─────────────────────┴──────┴───────────────┘

            Note that schema defaults are sourced from all MEDS schemas:

            >>> MEDSDataset.parse_csv(
            ...     'subject_id,time,code,numeric_value\\n0,"1/1/2025, 12:00:00",foo,1.0',
            ... )
            shape: (1, 4)
            ┌────────────┬─────────────────────┬──────┬───────────────┐
            │ subject_id ┆ time                ┆ code ┆ numeric_value │
            │ ---        ┆ ---                 ┆ ---  ┆ ---           │
            │ i64        ┆ datetime[μs]        ┆ str  ┆ f32           │
            ╞════════════╪═════════════════════╪══════╪═══════════════╡
            │ 0          ┆ 2025-01-01 12:00:00 ┆ foo  ┆ 1.0           │
            └────────────┴─────────────────────┴──────┴───────────────┘
            >>> MEDSDataset.parse_csv(
            ...     'code,description,parent_codes\\nfoo,foobar code,"bar,baz"',
            ... )
            shape: (1, 3)
            ┌──────┬─────────────┬──────────────┐
            │ code ┆ description ┆ parent_codes │
            │ ---  ┆ ---         ┆ ---          │
            │ str  ┆ str         ┆ list[str]    │
            ╞══════╪═════════════╪══════════════╡
            │ foo  ┆ foobar code ┆ ["bar,baz"]  │
            └──────┴─────────────┴──────────────┘
            >>> MEDSDataset.parse_csv(
            ...     'subject_id,split\\n0,train\\n1,test',
            ... )
            shape: (2, 2)
            ┌────────────┬───────┐
            │ subject_id ┆ split │
            │ ---        ┆ ---   │
            │ i64        ┆ str   │
            ╞════════════╪═══════╡
            │ 0          ┆ train │
            │ 1          ┆ test  │
            └────────────┴───────┘
            >>> MEDSDataset.parse_csv(
            ...     'subject_id,prediction_time,boolean_value\\n0,"1/1/2025, 12:00:00",False',
            ... )
            shape: (1, 3)
            ┌────────────┬─────────────────────┬───────────────┐
            │ subject_id ┆ prediction_time     ┆ boolean_value │
            │ ---        ┆ ---                 ┆ ---           │
            │ i64        ┆ datetime[μs]        ┆ bool          │
            ╞════════════╪═════════════════════╪═══════════════╡
            │ 0          ┆ 2025-01-01 12:00:00 ┆ false         │
            └────────────┴─────────────────────┴───────────────┘

            Note that columns are not verified to come from a single MEDS schema:

            >>> MEDSDataset.parse_csv(
            ...     'subject_id,prediction_time,split\\n0,"1/1/2025, 12:00:00",train',
            ... )
            shape: (1, 3)
            ┌────────────┬─────────────────────┬───────┐
            │ subject_id ┆ prediction_time     ┆ split │
            │ ---        ┆ ---                 ┆ ---   │
            │ i64        ┆ datetime[μs]        ┆ str   │
            ╞════════════╪═════════════════════╪═══════╡
            │ 0          ┆ 2025-01-01 12:00:00 ┆ train │
            └────────────┴─────────────────────┴───────┘

            Columns from MEDS schemas can also be overwritten with the schema updates keyword arguments:

            >>> MEDSDataset.parse_csv(
            ...     'subject_id,prediction_time,parent_codes\\n0,"1/1/2025, 12:00:00",train',
            ...     subject_id=pl.String, prediction_time=pl.String, parent_codes=pl.String
            ... )
            shape: (1, 3)
            ┌────────────┬────────────────────┬──────────────┐
            │ subject_id ┆ prediction_time    ┆ parent_codes │
            │ ---        ┆ ---                ┆ ---          │
            │ str        ┆ str                ┆ str          │
            ╞════════════╪════════════════════╪══════════════╡
            │ 0          ┆ 1/1/2025, 12:00:00 ┆ train        │
            └────────────┴────────────────────┴──────────────┘

            Columns can also be inferred from the provided CSV, though only for columns not in a dedicated
            MEDS schema:

            >>> MEDSDataset.parse_csv(
            ...     'subject_id,time,code_2,numeric_value\\n0,"1/1/2025, 12:00:00",foo,1.0',
            ... )
            shape: (1, 4)
            ┌────────────┬─────────────────────┬────────┬───────────────┐
            │ subject_id ┆ time                ┆ code_2 ┆ numeric_value │
            │ ---        ┆ ---                 ┆ ---    ┆ ---           │
            │ i64        ┆ datetime[μs]        ┆ str    ┆ f32           │
            ╞════════════╪═════════════════════╪════════╪═══════════════╡
            │ 0          ┆ 2025-01-01 12:00:00 ┆ foo    ┆ 1.0           │
            └────────────┴─────────────────────┴────────┴───────────────┘

            Errors are raised when the schema is incomplete, inaccurate, or the CSV is malformed:

            >>> MEDSDataset.parse_csv(
            ...     'subject_id,time,code,numeric_value\\n0,"1/1/2025, 12:00:00",foo,foo',
            ... )
            Traceback (most recent call last):
                ...
            ValueError: Failed to read:...
            >>> MEDSDataset.parse_csv(123)
            Traceback (most recent call last):
                ...
            ValueError: csv must be a string; got <class 'int'>
        """

        if not isinstance(csv, str):
            raise ValueError(f"csv must be a string; got {type(csv)}")

        read_schema = {}
        time_schema = {}
        has_parent_codes = False

        cols = csv.split("\n")[0].split(",")
        for col in cols:
            do_retype_time = col in cls.TIME_FIELDS
            do_retype_parent_codes = col == parent_codes_field
            if col in schema_updates:
                read_schema[col] = schema_updates[col]
                if col in cls.TIME_FIELDS and not schema_updates[col].is_temporal():
                    do_retype_time = False
                if col == parent_codes_field and schema_updates[col] is not cls.PL_CODE_METADATA_SCHEMA[col]:
                    do_retype_parent_codes = False
            elif col in cls.PL_DATA_SCHEMA:
                read_schema[col] = cls.PL_DATA_SCHEMA[col]
            elif col in cls.PL_CODE_METADATA_SCHEMA:
                read_schema[col] = cls.PL_CODE_METADATA_SCHEMA[col]
            elif col in cls.PL_SUBJECT_SPLIT_SCHEMA:
                read_schema[col] = cls.PL_SUBJECT_SPLIT_SCHEMA[col]
            elif col in cls.PL_LABEL_SCHEMA:
                read_schema[col] = cls.PL_LABEL_SCHEMA[col]
            else:
                logger.warning(f"Column {col} not found in schema")

            if do_retype_time:
                time_schema[col] = read_schema.pop(col)
                read_schema[col] = pl.String
            elif do_retype_parent_codes:
                has_parent_codes = True
                read_schema[col] = pl.String

        try:
            df = pl.read_csv(
                StringIO(csv), schema_overrides={col: read_schema[col] for col in cols if col in read_schema}
            )
        except Exception as e:
            raise ValueError(f"Failed to read:\n{csv}\nUnder schema:\n{read_schema}") from e

        col_updates = {t: pl.col(t).str.strptime(dt, cls.CSV_TS_FORMAT) for t, dt in time_schema.items()}
        if has_parent_codes:
            col_updates[parent_codes_field] = (
                pl.col(parent_codes_field).str.split(", ").cast(pl.List(pl.String))
            )

        return df.with_columns(**col_updates).select(cols)

    @classmethod
    def from_yaml(cls, yaml: str | Path, **schema_overrides) -> "MEDSDataset":
        """Create a MEDSDataset from a YAML string or file on disk.

        Args:
            yaml: The YAML string or file path to load the dataset from. This file should contain a flat set
                of string keys that correspond to file-paths relative to the MEDS-Root, with values that are
                strings of the associated data in CSV format (JSON format for dataset metadata). Missing keys
                corresponding to mandatory files will be inferred if possible or raise an error if not.

        Raises:
            ValueError: If the YAML is not a valid MEDSDataset.
            FileNotFoundError: If the file path does not exist.

        Returns:
            The MEDSDataset object reflected in the YAML file. If no code metadata is specified, an empty code
            metadata dataframe will be created. If no subject splits are specified, `None` will be returned.
            If no dataset metadata is specified, a default dataset metadata object will be created.

        Examples:
            >>> from meds_testing_helpers.static_sample_data import SIMPLE_STATIC_SHARDED_BY_SPLIT
            >>> D = MEDSDataset.from_yaml(SIMPLE_STATIC_SHARDED_BY_SPLIT)
            >>> print(D)
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - train/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[239684,239684,239684,239684,239684,...,1195293,1195293,1195293,1195293,1195293],[1195293]]
                time: [[null,null,1980-12-28 00:00:00.000000,2010-05-11 17:41:51.000000,2010-05-11 17:41:51.000000,...,2010-06-20 20:12:31.000000,2010-06-20 20:24:44.000000,2010-06-20 20:24:44.000000,2010-06-20 20:41:33.000000,2010-06-20 20:41:33.000000],[2010-06-20 20:50:04.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//CARDIAC","HR",...,"TEMP","HR","TEMP","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,175.27112,null,null,102.6,...,99.8,107.7,100,107.5,100.4],[null]]
              - train/1:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[68729,68729,68729,68729,68729,...,814703,814703,814703,814703,814703],[814703]]
                time: [[null,null,1978-03-09 00:00:00.000000,2010-05-26 02:30:56.000000,2010-05-26 02:30:56.000000,...,null,1976-03-28 00:00:00.000000,2010-02-05 05:55:39.000000,2010-02-05 05:55:39.000000,2010-02-05 05:55:39.000000],[2010-02-05 07:02:30.000000]]
                code: [["EYE_COLOR//HAZEL","HEIGHT","DOB","ADMISSION//PULMONARY","HR",...,"HEIGHT","DOB","ADMISSION//ORTHOPEDIC","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,160.39531,null,null,86,...,156.4856,null,null,170.2,100.1],[null]]
              - tuning/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[754281,754281,754281,754281,754281,754281],[754281]]
                time: [[null,null,1988-12-19 00:00:00.000000,2010-01-03 06:27:59.000000,2010-01-03 06:27:59.000000,2010-01-03 06:27:59.000000],[2010-01-03 08:22:13.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//PULMONARY","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,166.22261,null,null,142,99.8],[null]]
              - held_out/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733],[1500733]]
                time: [[null,null,1986-07-20 00:00:00.000000,2010-06-03 14:54:38.000000,2010-06-03 14:54:38.000000,2010-06-03 14:54:38.000000,2010-06-03 15:39:49.000000,2010-06-03 15:39:49.000000,2010-06-03 16:20:49.000000,2010-06-03 16:20:49.000000],[2010-06-03 16:44:26.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//ORTHOPEDIC","HR","TEMP","HR","TEMP","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,158.60132,null,null,91.4,100,84.4,100.3,90.1,100.1],[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: [["EYE_COLOR//BLUE","EYE_COLOR//BROWN","EYE_COLOR//HAZEL","HR"],["TEMP"]]
              description: [["Blue Eyes. Less common than brown.","Brown Eyes. The most common eye color.","Hazel eyes. These are uncommon","Heart Rate"],["Body Temperature"]]
              parent_codes: [[null,null,null,["LOINC/8867-4"]],[["LOINC/8310-5"]]]
            subject_splits:
              pyarrow.Table
              subject_id: int64
              split: string
              ----
              subject_id: [[239684,1195293,68729,814703,754281],[1500733]]
              split: [["train","train","train","train","tuning"],["held_out"]]

            You can also read from a filepath directly:

            >>> import tempfile
            >>> yaml_lines = [
            ...    "data/train/0: |-2",
            ...    "  subject_id,time,code,numeric_value",
            ...    '  0,"1/1/2025, 12:00:00",A,',
            ...    "metadata/subject_splits.parquet: |-2",
            ...    "  subject_id,split",
            ...    "  0,train",
            ...    "metadata/dataset.json:",
            ...    "  dataset_name: test",
            ...    "  dataset_version: 0.0.1",
            ... ]
            >>> with tempfile.NamedTemporaryFile("w", suffix=".yaml") as f:
            ...     for line in yaml_lines:
            ...         _ = f.write(f"{line}\\n")
            ...     _ = f.flush()
            ...     D = MEDSDataset.from_yaml(f.name)
            ...     print(repr(D)) # doctest: +NORMALIZE_WHITESPACE
            MEDSDataset(data_shards={'train/0': {'subject_id': [0],
                                                 'time': [datetime.datetime(2025, 1, 1, 12, 0)],
                                                 'code': ['A'], 'numeric_value': [None]}},
                        dataset_metadata={'dataset_name': 'test',
                                          'dataset_version': '0.0.1'},
                        code_metadata={'code': [], 'description': [], 'parent_codes': []},
                        subject_splits={'subject_id': [0], 'split': ['train']})

            Though task labels are not formalized in MEDS (in terms of storage on disk; see
            https://github.com/Medical-Event-Data-Standard/meds/issues/75 for more information), you can also
            track a collection of sharded task labels by task name in this class:

            >>> from meds_testing_helpers.static_sample_data import SIMPLE_STATIC_SHARDED_BY_SPLIT_WITH_TASKS
            >>> D = MEDSDataset.from_yaml(SIMPLE_STATIC_SHARDED_BY_SPLIT_WITH_TASKS)
            >>> print(D)
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - train/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[239684,239684,239684,239684,239684,...,1195293,1195293,1195293,1195293,1195293],[1195293]]
                time: [[null,null,1980-12-28 00:00:00.000000,2010-05-11 17:41:51.000000,2010-05-11 17:41:51.000000,...,2010-06-20 20:12:31.000000,2010-06-20 20:24:44.000000,2010-06-20 20:24:44.000000,2010-06-20 20:41:33.000000,2010-06-20 20:41:33.000000],[2010-06-20 20:50:04.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//CARDIAC","HR",...,"TEMP","HR","TEMP","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,175.27112,null,null,102.6,...,99.8,107.7,100,107.5,100.4],[null]]
              - train/1:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[68729,68729,68729,68729,68729,...,814703,814703,814703,814703,814703],[814703]]
                time: [[null,null,1978-03-09 00:00:00.000000,2010-05-26 02:30:56.000000,2010-05-26 02:30:56.000000,...,null,1976-03-28 00:00:00.000000,2010-02-05 05:55:39.000000,2010-02-05 05:55:39.000000,2010-02-05 05:55:39.000000],[2010-02-05 07:02:30.000000]]
                code: [["EYE_COLOR//HAZEL","HEIGHT","DOB","ADMISSION//PULMONARY","HR",...,"HEIGHT","DOB","ADMISSION//ORTHOPEDIC","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,160.39531,null,null,86,...,156.4856,null,null,170.2,100.1],[null]]
              - tuning/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[754281,754281,754281,754281,754281,754281],[754281]]
                time: [[null,null,1988-12-19 00:00:00.000000,2010-01-03 06:27:59.000000,2010-01-03 06:27:59.000000,2010-01-03 06:27:59.000000],[2010-01-03 08:22:13.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//PULMONARY","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,166.22261,null,null,142,99.8],[null]]
              - held_out/0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733,1500733],[1500733]]
                time: [[null,null,1986-07-20 00:00:00.000000,2010-06-03 14:54:38.000000,2010-06-03 14:54:38.000000,2010-06-03 14:54:38.000000,2010-06-03 15:39:49.000000,2010-06-03 15:39:49.000000,2010-06-03 16:20:49.000000,2010-06-03 16:20:49.000000],[2010-06-03 16:44:26.000000]]
                code: [["EYE_COLOR//BROWN","HEIGHT","DOB","ADMISSION//ORTHOPEDIC","HR","TEMP","HR","TEMP","HR","TEMP"],["DISCHARGE"]]
                numeric_value: [[null,158.60132,null,null,91.4,100,84.4,100.3,90.1,100.1],[null]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: [["EYE_COLOR//BLUE","EYE_COLOR//BROWN","EYE_COLOR//HAZEL","HR"],["TEMP"]]
              description: [["Blue Eyes. Less common than brown.","Brown Eyes. The most common eye color.","Hazel eyes. These are uncommon","Heart Rate"],["Body Temperature"]]
              parent_codes: [[null,null,null,["LOINC/8867-4"]],[["LOINC/8310-5"]]]
            subject_splits:
              pyarrow.Table
              subject_id: int64
              split: string
              ----
              subject_id: [[239684,1195293,68729,814703,754281],[1500733]]
              split: [["train","train","train","train","tuning"],["held_out"]]
            task labels:
              * boolean_value_task:
                - labels_A.parquet:
                  pyarrow.Table
                  subject_id: int64
                  prediction_time: timestamp[us]
                  boolean_value: bool
                  integer_value: int64
                  float_value: double
                  categorical_value: string
                  ----
                  subject_id: [[239684,239684,239684,1195293,1195293,1195293,68729,68729,68729],[68729]]
                  prediction_time: [[2010-05-11 18:00:00.000000,2010-05-11 18:30:00.000000,2010-05-11 19:00:00.000000,2010-06-20 19:30:00.000000,2010-06-20 20:00:00.000000,2010-06-20 20:30:00.000000,2010-05-26 03:00:00.000000,2010-05-26 03:30:00.000000,2010-05-26 04:00:00.000000],[2010-05-26 04:30:00.000000]]
                  boolean_value: [[false,true,true,false,true,true,false,false,true],[true]]
                  integer_value: [[null,null,null,null,null,null,null,null,null],[null]]
                  float_value: [[null,null,null,null,null,null,null,null,null],[null]]
                  categorical_value: [[null,null,null,null,null,null,null,null,null],[null]]
                - labels_B.parquet:
                  pyarrow.Table
                  subject_id: int64
                  prediction_time: timestamp[us]
                  boolean_value: bool
                  integer_value: int64
                  float_value: double
                  categorical_value: string
                  ----
                  subject_id: [[814703,814703,814703,754281,754281,754281,754281,1500733,1500733,1500733],[1500733]]
                  prediction_time: [[2010-02-05 06:00:00.000000,2010-02-05 06:30:00.000000,2010-02-05 07:00:00.000000,2010-01-03 06:30:00.000000,2010-01-03 07:00:00.000000,2010-01-03 07:30:00.000000,2010-01-03 08:00:00.000000,2010-06-03 15:00:00.000000,2010-06-03 15:30:00.000000,2010-06-03 16:00:00.000000],[2010-06-03 16:30:00.000000]]
                  boolean_value: [[false,true,true,false,false,true,true,false,false,true],[true]]
                  integer_value: [[null,null,null,null,null,null,null,null,null,null],[null]]
                  float_value: [[null,null,null,null,null,null,null,null,null,null],[null]]
                  categorical_value: [[null,null,null,null,null,null,null,null,null,null],[null]]

            Errors are raised when the YAML is malformed or a non-existent path:

            >>> MEDSDataset.from_yaml(123)
            Traceback (most recent call last):
                ...
            ValueError: yaml must be a string or a file path; got <class 'int'>
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            ...     MEDSDataset.from_yaml(Path(tmpdir) / "nonexistent.yaml")
            Traceback (most recent call last):
                ...
            FileNotFoundError: File not found: ...
            >>> MEDSDataset.from_yaml("foo: bar")
            Traceback (most recent call last):
                ...
            ValueError: Unrecognized key in YAML: foo. Must start with 'data/' or 'metadata/'.
            >>> MEDSDataset.from_yaml("metadata/codes.parquet: [1, 2]")
            Traceback (most recent call last):
                ...
            ValueError: Expected value for key metadata/codes.parquet to be a string, got <class 'list'>
            >>> MEDSDataset.from_yaml("metadata/foo: bar")
            Traceback (most recent call last):
                ...
            ValueError: Unrecognized key in YAML: metadata/foo
            >>> MEDSDataset.from_yaml("metadata/dataset.json: {dataset_name: test, dataset_version: 0.0.1}")
            Traceback (most recent call last):
                ...
            ValueError: No data shards found in YAML
            >>> MEDSDataset.from_yaml('metadata/dataset.json: "{dataset_name: test, dataset_version: 0.0.1}"')
            Traceback (most recent call last):
                ...
            ValueError: Expected value for key metadata/dataset.json to be a dict, got <class 'str'>
        """  # noqa: E501
        if isinstance(yaml, str) and yaml.endswith(".yaml"):
            logger.debug(f"Inferring yaml {yaml} is a file path as it ends with '.yaml'")
            yaml = Path(yaml)

        match yaml:
            case Path() if yaml.is_file():
                yaml = yaml.read_text().strip()
            case Path():
                raise FileNotFoundError(f"File not found: {yaml}")
            case str():
                yaml = yaml.strip()
            case _:
                raise ValueError(f"yaml must be a string or a file path; got {type(yaml)}")

        data = load_yaml(yaml, Loader=Loader)

        data_shards = {}
        code_metadata = None
        subject_splits = None
        dataset_metadata = DatasetMetadata()
        task_labels = None
        for key, value in data.items():
            key_parts = key.split("/")

            if key == cls.TASK_LABELS_SUBDIR:
                pass
            elif len(key_parts) < 2 or key_parts[0] not in {data_subdirectory, "metadata"}:
                raise ValueError(f"Unrecognized key in YAML: {key}. Must start with 'data/' or 'metadata/'.")

            if key in {dataset_metadata_filepath, cls.TASK_LABELS_SUBDIR}:
                if not isinstance(value, dict):
                    raise ValueError(f"Expected value for key {key} to be a dict, got {type(value)}")
            elif not isinstance(value, str):
                raise ValueError(f"Expected value for key {key} to be a string, got {type(value)}")

            root = key_parts[0]

            if root == data_subdirectory:
                rest = "/".join(key_parts[1:])
                data_shards[rest.replace(".parquet", "")] = cls.parse_csv(value, **schema_overrides)
            elif key == code_metadata_filepath:
                code_metadata = cls.parse_csv(value, **schema_overrides)
            elif key == subject_splits_filepath:
                subject_splits = cls.parse_csv(value, **schema_overrides)
            elif key == dataset_metadata_filepath:
                dataset_metadata = DatasetMetadata(**value, **schema_overrides)
            elif key == cls.TASK_LABELS_SUBDIR:
                task_labels = {
                    task_name: {
                        shard: cls.parse_csv(data, **schema_overrides) for shard, data in shards.items()
                    }
                    for task_name, shards in value.items()
                }
            else:
                raise ValueError(f"Unrecognized key in YAML: {key}")

        if len(data_shards) == 0:
            raise ValueError("No data shards found in YAML")

        return cls(
            data_shards=data_shards,
            dataset_metadata=dataset_metadata,
            code_metadata=code_metadata,
            subject_splits=subject_splits,
            task_labels=task_labels,
        )

    @staticmethod
    def _align_df(df: pl.DataFrame, schema: pa.Schema) -> pa.Table:
        return df.select(schema.names).to_arrow().cast(schema)

    @property
    def dataset_metadata_fp(self) -> Path | None:
        if self.root_dir is None:
            return None
        else:
            return self.root_dir / dataset_metadata_filepath

    @property
    def dataset_metadata(self) -> DatasetMetadata:
        if self.root_dir is None:
            return self._dataset_metadata
        else:
            return DatasetMetadata(**json.loads(self.dataset_metadata_fp.read_text()))

    @dataset_metadata.setter
    def dataset_metadata(self, value: DatasetMetadata | None):
        self._dataset_metadata = value

    def _shard_name(self, data_fp: Path, root_dir: Path | None = None) -> str:
        if root_dir is None:
            root_dir = self.root_dir / data_subdirectory
        return data_fp.relative_to(root_dir).with_suffix("").as_posix()

    @property
    def shard_fps(self) -> list[Path] | None:
        if self.root_dir is None:
            return None
        else:
            return sorted(list((self.root_dir / data_subdirectory).rglob("*.parquet")))

    @property
    def _pl_shards(self) -> SHARDED_DF_T:
        if self._data_shards is None:
            return {self._shard_name(fp): pl.read_parquet(fp, use_pyarrow=True) for fp in self.shard_fps}
        else:
            return self._data_shards

    @property
    def data_shards(self) -> dict[str, pa.Table]:
        return {shard: self._align_df(df, data_schema()) for shard, df in self._pl_shards.items()}

    @data_shards.setter
    def data_shards(self, value: SHARDED_DF_T | None):
        self._data_shards = value

    @property
    def code_metadata_fp(self) -> Path | None:
        if self.root_dir is None:
            return None
        else:
            return self.root_dir / code_metadata_filepath

    @property
    def _pl_code_metadata(self) -> pl.DataFrame:
        if self._code_metadata is None:
            return pl.read_parquet(self.code_metadata_fp, use_pyarrow=True)
        else:
            return self._code_metadata

    @property
    def code_metadata(self) -> pa.Table:
        return self._align_df(self._pl_code_metadata, code_metadata_schema())

    @code_metadata.setter
    def code_metadata(self, value: pl.DataFrame | None):
        self._code_metadata = value

    @property
    def subject_splits_fp(self) -> Path | None:
        if self.root_dir is None:
            return None
        else:
            return self.root_dir / subject_splits_filepath

    @property
    def _pl_subject_splits(self) -> pl.DataFrame:
        if self.root_dir is None:
            return self._subject_splits

        if self.subject_splits_fp.exists():
            return pl.read_parquet(self.subject_splits_fp, use_pyarrow=True)
        else:
            return None

    @property
    def subject_splits(self) -> pa.Table | None:
        pl_subject_splits = self._pl_subject_splits
        if pl_subject_splits is None:
            return None
        return self._align_df(pl_subject_splits, subject_split_schema)

    @subject_splits.setter
    def subject_splits(self, value: pl.DataFrame | None):
        self._subject_splits = value

    @property
    def task_root_dir(self) -> Path | None:
        if self.root_dir is None:
            return None
        else:
            return self.root_dir / self.TASK_LABELS_SUBDIR

    @property
    def task_names_fp(self) -> Path | None:
        if self.task_root_dir is None:
            return None
        else:
            return self.task_root_dir / self.TASK_NAMES_FN

    @property
    def task_label_fps(self) -> dict[str, list[Path]] | None:
        if self.root_dir is None:
            return None

        if not self.task_root_dir.is_dir():
            return None
        else:
            out = {}
            task_names = json.loads(self.task_names_fp.read_text())
            for task in task_names:
                task_dir = self.task_root_dir / task
                out[task] = sorted(list(task_dir.rglob("*.parquet")))
            return out

    @property
    def _pl_task_labels(self) -> dict[str, SHARDED_DF_T] | None:
        if self.root_dir is None:
            return self._task_labels
        elif self.task_label_fps is None:
            return None
        else:
            out = {}
            for task, shard_fps in self.task_label_fps.items():
                out[task] = {
                    self._shard_name(fp, self.task_root_dir / task): pl.read_parquet(fp, use_pyarrow=True)
                    for fp in shard_fps
                }
            return out

    @property
    def task_labels(self) -> dict[str, pa.Table]:
        if self._pl_task_labels is None:
            return None

        return {
            task_name: {shard: self._align_df(df, label_schema) for shard, df in shards.items()}
            for task_name, shards in self._pl_task_labels.items()
        }

    @task_labels.setter
    def task_labels(self, value: dict[str, SHARDED_DF_T] | None):
        self._task_labels = value

    @property
    def task_names(self) -> list[str] | None:
        if self.task_labels is None:
            return None
        else:
            return list(self.task_labels.keys())

    def write(self, output_dir: Path) -> "MEDSDataset":
        data_dir = output_dir / data_subdirectory

        for shard, table in self.data_shards.items():
            fp = data_dir / f"{shard}.parquet"
            fp.parent.mkdir(parents=True, exist_ok=True)
            pq.write_table(table, fp)

        code_metadata_fp = output_dir / code_metadata_filepath
        code_metadata_fp.parent.mkdir(parents=True, exist_ok=True)
        pq.write_table(self.code_metadata, code_metadata_fp)

        dataset_metadata_fp = output_dir / dataset_metadata_filepath
        dataset_metadata_fp.parent.mkdir(parents=True, exist_ok=True)
        dataset_metadata_fp.write_text(json.dumps(self.dataset_metadata))

        if self.subject_splits is not None:
            subject_splits_fp = output_dir / subject_splits_filepath
            subject_splits_fp.parent.mkdir(parents=True, exist_ok=True)
            pq.write_table(self.subject_splits, subject_splits_fp)

        if self.task_labels:
            task_labels_dir = output_dir / self.TASK_LABELS_SUBDIR
            task_labels_dir.mkdir(exist_ok=True, parents=True)

            task_names_fp = task_labels_dir / self.TASK_NAMES_FN
            task_names_fp.write_text(json.dumps(self.task_names))

            for task_name, shards in self.task_labels.items():
                for shard, table in shards.items():
                    fp = task_labels_dir / task_name / f"{shard}.parquet"
                    fp.parent.mkdir(parents=True, exist_ok=True)
                    pq.write_table(table, fp)

        return MEDSDataset(root_dir=output_dir)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        if self.root_dir is None:
            kwargs = {
                "data_shards": {k: v.to_dict(as_series=False) for k, v in self._pl_shards.items()},
                "dataset_metadata": self.dataset_metadata,
                "code_metadata": self._pl_code_metadata.to_dict(as_series=False),
            }
            if self.subject_splits is not None:
                kwargs["subject_splits"] = self._pl_subject_splits.to_dict(as_series=False)
            if self.task_labels is not None:
                kwargs["task_labels"] = {
                    task_name: {shard: df.to_dict(as_series=False) for shard, df in shards.items()}
                    for task_name, shards in self._pl_task_labels.items()
                }
            kwargs_str = ", ".join(f"{k}={repr(v)}" for k, v in kwargs.items())
            return f"{cls_name}({kwargs_str})"
        else:
            return f"{cls_name}(root_dir={repr(self.root_dir)})"

    def __str__(self) -> str:
        lines = []
        lines.append(f"{self.__class__.__name__}:")
        if self.root_dir is not None:
            lines.append(f"stored in root_dir: {str(self.root_dir.resolve())}")
        lines.append("dataset_metadata:")
        for k, v in self.dataset_metadata.items():
            lines.append(f"  - {k}: {v}")
        lines.append("data_shards:")
        for shard, table in self.data_shards.items():
            lines.append(f"  - {shard}:")
            lines.append("    " + str(table).replace("\n", "\n    "))
        lines.append("code_metadata:")
        lines.append("  " + str(self.code_metadata).replace("\n", "\n  "))
        if self.subject_splits is None:
            lines.append("subject_splits: None")
        else:
            lines.append("subject_splits:")
            lines.append("  " + str(self.subject_splits).replace("\n", "\n  "))
        if self.task_labels is not None:
            lines.append("task labels:")
            for task_name, shards in self.task_labels.items():
                lines.append(f"  * {task_name}:")
                for shard, table in shards.items():
                    lines.append(f"    - {shard}:")
                    lines.append("      " + str(table).replace("\n", "\n      "))

        return "\n".join(lines)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MEDSDataset):
            return False

        if self.data_shards != other.data_shards:
            return False
        if self.dataset_metadata != other.dataset_metadata:
            return False
        if self.code_metadata != other.code_metadata:
            return False
        if self.subject_splits != other.subject_splits:
            return False
        if self.task_labels != other.task_labels:
            return False

        return True



"""Pytest plugin capabilities for loading sample MEDS datasets."""

import subprocess
import tempfile
from pathlib import Path

import pytest

from .dataset import MEDSDataset
from .static_sample_data import (
    SIMPLE_STATIC_SHARDED_BY_SPLIT,
    SIMPLE_STATIC_SHARDED_BY_SPLIT_WITH_TASKS,
)


def pytest_addoption(parser):  # pragma: no cover
    parser.addoption(
        "--generated-dataset-N",
        type=int,
        help="If using the generated-dataset dataset, how many patients?",
        default=500,
    )

    parser.addoption(
        "--generated-dataset-seed",
        type=int,
        help="If generating a dataset, what seed to use?",
        default=1,
    )

    parser.addoption(
        "--MEDS-datasets-scope",
        type=str,
        help="For generated or static dataset fixtures, what scope to use?",
        default="session",
    )


def get_MEDS_datasets_scope(fixture_name: str, config) -> str:
    return config.getoption("--MEDS-datasets-scope")


@pytest.fixture(scope=get_MEDS_datasets_scope)
def simple_static_MEDS() -> Path:
    with tempfile.TemporaryDirectory() as data_dir:
        data_dir = Path(data_dir)
        data = MEDSDataset.from_yaml(SIMPLE_STATIC_SHARDED_BY_SPLIT)
        data.write(data_dir)
        yield data_dir


@pytest.fixture(scope=get_MEDS_datasets_scope)
def simple_static_MEDS_dataset_with_task() -> Path:
    with tempfile.TemporaryDirectory() as data_dir:
        data_dir = Path(data_dir)
        data = MEDSDataset.from_yaml(SIMPLE_STATIC_SHARDED_BY_SPLIT_WITH_TASKS)
        data.write(data_dir)
        yield data_dir


def generate_MEDS(request, dataset_spec: str) -> Path:
    N = request.config.getoption("--generated-dataset-N")
    seed = request.config.getoption("--generated-dataset-seed")
    with tempfile.TemporaryDirectory() as data_dir:
        data_dir = Path(data_dir)

        cmd_args = [
            "build_sample_MEDS_dataset",
            f"seed={seed}",
            f"N_subjects={N}",
            "do_overwrite=False",
            f"output_dir={str(data_dir)}",
            "dataset_spec/data_generator=sample",
        ]

        out = subprocess.run(cmd_args, shell=False, check=False, capture_output=True)

        error_str = (
            f"Command failed with return code {out.returncode}.\n"
            f"Command stdout:\n{out.stdout.decode()}\n"
            f"Command stderr:\n{out.stderr.decode()}"
        )

        if out.returncode != 0:  # pragma: no cover
            raise RuntimeError(error_str)

        yield data_dir


@pytest.fixture(scope=get_MEDS_datasets_scope)
def generated_sample_MEDS(request) -> Path:
    yield from generate_MEDS(request, "sample")



from importlib.resources import files

from .. import __package_name__

root = files(__package_name__).joinpath("static_sample_data")

static_dataset_yamls = root.rglob("*.yaml")

exported_yamls = {}
for yaml_path in static_dataset_yamls:
    yaml_name = yaml_path.relative_to(root).with_suffix("").as_posix()

    nested = [x.upper() for x in yaml_name.split("/")]
    current = exported_yamls
    for nested_key in nested[:-1]:
        current = current.setdefault(nested_key, {})
    current[nested[-1]] = yaml_path.read_text()

for n, val in exported_yamls.items():
    globals()[n] = val

__all__ = list(exported_yamls.keys()) + ["exported_yamls"]



import subprocess
import tempfile
from pathlib import Path

from meds_testing_helpers.dataset import MEDSDataset


def test_dataset_generation(generated_sample_MEDS):
    """Test the generation of a dataset."""

    # This will throw an error if the generated data is invalid
    MEDSDataset(root_dir=generated_sample_MEDS)


def test_error_cases():
    cmd_args = [
        "build_sample_MEDS_dataset",
        "seed=1",
        "N_subjects=10",
        "dataset_spec/data_generator=sample",
    ]
    with tempfile.TemporaryDirectory() as data_dir:
        output_is_file = Path(data_dir) / "output_is_file"
        output_is_file.touch()

        out = subprocess.run(
            cmd_args + [f"output_dir={str(output_is_file)}"], shell=False, check=False, capture_output=True
        )
        assert out.returncode != 0, "Should fail because output is a file"

        output_is_existing = Path(data_dir) / "output_is_existing"
        out_data = output_is_existing / "data"

        out_data.mkdir(parents=True)

        (out_data / "data1.parquet").touch()
        out = subprocess.run(
            cmd_args + [f"output_dir={str(output_is_existing)}"],
            shell=False,
            check=False,
            capture_output=True,
        )
        assert out.returncode != 0, "Should fail because output data subdir exists"

        out = subprocess.run(
            cmd_args + [f"output_dir={str(output_is_existing)}", "do_overwrite=True"],
            shell=False,
            check=False,
            capture_output=True,
        )
        assert out.returncode == 0, "Should succeed because output data subdir exists but overwrite is True"



#!/usr/bin/env python

import dataclasses
import json
import logging
from pathlib import Path

import hydra
import numpy as np
import polars as pl
from meds import (
    birth_code,
    code_field,
    dataset_metadata_filepath,
    death_code,
    held_out_split,
    subject_id_field,
    subject_splits_filepath,
    time_field,
    train_split,
    tuning_split,
)
from omegaconf import DictConfig, OmegaConf

from . import INF_YAML
from .dataset_generator import (
    DatetimeGenerator,
    MEDSDataDFGenerator,
    MEDSDatasetGenerator,
    PositiveIntGenerator,
    PositiveTimeDeltaGenerator,
    ProportionGenerator,
)

logger = logging.getLogger(__name__)


def to_dt_list(df: pl.DataFrame, col: str) -> list[str]:
    """Converts a Polars DataFrame column to a list of ISO format strings.

    Args:
        df: The DataFrame.
        col: The column name.

    Returns:
        A list of ISO format strings.

    Examples:
        >>> from datetime import datetime
        >>> df = pl.DataFrame({"col": [datetime(2021, 1, 1), datetime(2021, 1, 2)]})
        >>> to_dt_list(df, "col")
        ['2021-01-01T00:00:00', '2021-01-02T00:00:00']
    """
    return df.select(pl.col(col).drop_nulls().dt.strftime("%Y-%m-%dT%H:%M:%S%.f"))[col].to_list()


@hydra.main(version_base=None, config_path=str(INF_YAML.parent), config_name=INF_YAML.stem)
def main(cfg: DictConfig):
    """Infers the configuration parameters that would generate a dataset similar to the input."""

    output_fp = Path(cfg.output_fp)
    if output_fp.exists():
        if cfg.do_overwrite:
            logger.info(f"Overwriting existing file {output_fp}.")
        else:
            raise FileExistsError(f"Output file {output_fp} already exists.")

    data_dir = Path(cfg.dataset_dir) / "data"
    metadata_dir = Path(cfg.dataset_dir) / "metadata"

    shards = list(data_dir.rglob("*.parquet"))
    if not shards:
        raise FileNotFoundError(f"No shards found in {data_dir}")

    dataset_metadata_fp = metadata_dir / dataset_metadata_filepath
    if not dataset_metadata_fp.exists():
        logger.warning(f"Dataset metadata file not found at {dataset_metadata_fp}!")
        dataset_name = "UNKNOWN"
    else:
        dataset_metadata = json.loads(dataset_metadata_fp.read_text())
        dataset_name = f"{dataset_metadata['dataset_name']}/SYNTHETIC"

    subject_splits_fp = metadata_dir / subject_splits_filepath
    if not subject_splits_fp.exists():
        train_frac = None
        tuning_frac = None
        held_out_frac = None
    else:
        subject_splits = pl.read_parquet(subject_splits_fp)
        split_cnts = subject_splits.group_by("split").agg(pl.col(subject_id_field).count().alias("count"))
        train_cnt = split_cnts.filter(pl.col("split") == train_split).select("count").first().item()
        tuning_cnt = split_cnts.filter(pl.col("split") == tuning_split).select("count").first().item()
        held_out_cnt = split_cnts.filter(pl.col("split") == held_out_split).select("count").first().item()

        total_cnt = train_cnt + tuning_cnt + held_out_cnt
        train_frac = train_cnt / total_cnt
        tuning_frac = tuning_cnt / total_cnt
        held_out_frac = 1 - train_frac - tuning_frac

    if cfg.seed is not None:
        rng = np.random.default_rng(cfg.seed)
    else:
        rng = np.random.default_rng()

    rng.shuffle(shards)
    shards_to_examine = shards[:3]

    code_col = pl.col(code_field)
    time_col = pl.col(time_field)
    is_static = time_col.is_null()
    is_dynamic = time_col.is_not_null()
    is_birth = code_col.str.starts_with(birth_code)
    is_death = code_col.str.starts_with(death_code)
    is_dynamic_data = is_dynamic & ~is_birth & ~is_death
    numerics_present = (
        pl.col("numeric_value").is_not_null()
        & pl.col("numeric_value").is_finite()
        & pl.col("numeric_value").is_not_nan()
    )

    shards = [pl.read_parquet(shard, use_pyarrow=True) for shard in shards_to_examine]
    shard_sizes = [shard.select(pl.col(subject_id_field).n_unique()).item() for shard in shards]

    df = pl.concat(shards, how="vertical_relaxed")
    dynamic_df = df.filter(is_dynamic_data)

    static_vocab_size = df.filter(is_static).select(code_col.n_unique()).item()
    dynamic_vocab_size = dynamic_df.select(code_col.n_unique()).item()
    birth_codes_vocab_size = df.filter(is_birth).select(code_col.n_unique()).item()
    death_codes_vocab_size = df.filter(is_death).select(code_col.n_unique()).item()

    subject_stats = df.group_by(subject_id_field).agg(
        pl.when(is_birth).then(time_col).min().alias("birth_time"),
        pl.when(is_death).then(time_col).max().alias("death_time"),
        is_birth.any().alias("has_birth"),
        is_death.any().alias("has_death"),
        is_static.sum().alias("n_static_measurements"),
    )
    birth_times = to_dt_list(subject_stats, "birth_time")

    frac_subjects_with_birth = subject_stats.select(pl.col("has_birth").mean()).item()
    frac_subjects_with_death = subject_stats.select(pl.col("has_death").mean()).item()
    num_static_measurements_per_subject = subject_stats["n_static_measurements"].drop_nulls().to_list()

    subject_dynamic_stats = dynamic_df.group_by(subject_id_field).agg(
        time_col.min().alias("first_data_time"),
        time_col.max().alias("last_data_time"),
        time_col.n_unique().alias("n_events"),
        time_col.unique(maintain_order=True).diff().mean().alias("time_between_events"),
    )

    start_of_data_times = to_dt_list(subject_dynamic_stats, "first_data_time")
    num_events_per_subject = subject_dynamic_stats["n_events"].drop_nulls().to_list()
    time_between_data_events = subject_dynamic_stats["time_between_events"].drop_nulls().to_numpy()

    boundary_deltas = subject_dynamic_stats.join(subject_stats, on=subject_id_field).select(
        (pl.col("first_data_time") - pl.col("birth_time")).alias("time_between_birth_and_data"),
        (pl.col("death_time") - pl.col("last_data_time")).alias("time_between_data_and_death"),
    )

    time_between_birth_and_data = boundary_deltas["time_between_birth_and_data"].drop_nulls().to_numpy()
    time_between_data_and_death = boundary_deltas["time_between_data_and_death"].drop_nulls().to_numpy()

    num_measurements_per_event = (
        dynamic_df.group_by(subject_id_field, time_col).agg(pl.count())["count"].to_list()
    )

    dynamic_code_stats = (
        dynamic_df.group_by(code_field)
        .agg(numerics_present.sum().alias("n_values"), pl.count().alias("n_occurrences"))
        .select((pl.col("n_values") / pl.col("n_occurrences")).alias("frac_values"))
    )
    frac_dynamic_code_occurrences_with_values = dynamic_code_stats["frac_values"].to_list()

    static_code_stats = (
        df.filter(is_static)
        .group_by(code_field)
        .agg(numerics_present.sum().alias("n_values"), pl.count().alias("n_occurrences"))
        .select((pl.col("n_values") / pl.col("n_occurrences")).alias("frac_values"))
    )
    frac_static_code_occurrences_with_values = static_code_stats["frac_values"].to_list()

    birth_kwargs = {}
    if len(birth_times) > 0:
        birth_kwargs["birth_datetime_per_subject"] = DatetimeGenerator(birth_times)
        birth_kwargs["birth_codes_vocab_size"] = birth_codes_vocab_size
        birth_kwargs["time_between_birth_and_data_per_subject"] = PositiveTimeDeltaGenerator(
            time_between_birth_and_data
        )
    else:
        birth_kwargs["birth_datetime_per_subject"] = None
        birth_kwargs["birth_codes_vocab_size"] = 0
        birth_kwargs["time_between_birth_and_data_per_subject"] = None

    death_kwargs = {}
    if len(time_between_data_and_death) > 0:
        death_kwargs["death_codes_vocab_size"] = death_codes_vocab_size
        death_kwargs["time_between_data_and_death_per_subject"] = PositiveTimeDeltaGenerator(
            time_between_data_and_death
        )
    else:
        death_kwargs["death_codes_vocab_size"] = 0
        death_kwargs["time_between_data_and_death_per_subject"] = None

    data_generator = MEDSDataDFGenerator(
        start_data_datetime_per_subject=DatetimeGenerator(start_of_data_times),
        time_between_data_events_per_subject=PositiveTimeDeltaGenerator(time_between_data_events),
        num_events_per_subject=PositiveIntGenerator(num_events_per_subject),
        num_measurements_per_event=PositiveIntGenerator(num_measurements_per_event),
        num_static_measurements_per_subject=PositiveIntGenerator(num_static_measurements_per_subject),
        frac_dynamic_code_occurrences_with_value=ProportionGenerator(
            frac_dynamic_code_occurrences_with_values
        ),
        frac_static_code_occurrences_with_value=ProportionGenerator(frac_static_code_occurrences_with_values),
        static_vocab_size=static_vocab_size,
        dynamic_vocab_size=dynamic_vocab_size,
        frac_subjects_with_death=frac_subjects_with_death,
        frac_subjects_with_birth=frac_subjects_with_birth,
        **birth_kwargs,
        **death_kwargs,
    )

    dataset_generator = MEDSDatasetGenerator(
        data_generator=data_generator,
        shard_size=int(np.median(shard_sizes)),
        train_frac=train_frac,
        tuning_frac=tuning_frac,
        held_out_frac=held_out_frac,
        dataset_name=dataset_name,
    )

    dataset_config = DictConfig(dataclasses.asdict(dataset_generator))
    OmegaConf.save(dataset_config, output_fp)



from typing import Annotated, Any

import numpy as np
from annotated_types import Ge, Gt, Le

NUM = int | float
POSITIVE_NUM = Annotated[NUM, Gt(0)]
POSITIVE_INT = Annotated[int, Gt(0)]
NON_NEGATIVE_NUM = Annotated[NUM, Ge(0)]
NON_NEGATIVE_INT = Annotated[int, Ge(0)]
PROPORTION = Annotated[NUM, Ge(0), Le(1)]
POSITIVE_TIMEDELTA = Annotated[np.timedelta64, Gt(0)]


def is_NUM(x: Any) -> bool:
    """Check if x is a number.

    Examples:
        >>> is_NUM(1)
        True
        >>> is_NUM(1.0)
        True
        >>> is_NUM("a")
        False
    """
    return isinstance(x, (int, float))


def is_POSITIVE_NUM(x: Any) -> bool:
    """Check if x is a positive number.

    Examples:
        >>> is_POSITIVE_NUM(1)
        True
        >>> is_POSITIVE_NUM(0)
        False
        >>> is_POSITIVE_NUM(1.0)
        True
        >>> is_POSITIVE_NUM("foo")
        False
    """
    return is_NUM(x) and x > 0


def is_POSITIVE_INT(x: Any) -> bool:
    """Check if x is a positive integer.

    Examples:
        >>> is_POSITIVE_INT(1)
        True
        >>> is_POSITIVE_INT(0)
        False
        >>> is_POSITIVE_INT(1.0)
        False
    """
    return isinstance(x, int) and x > 0


def is_NON_NEGATIVE_NUM(x: Any) -> bool:
    """Check if x is a non-negative number.

    Examples:
        >>> is_NON_NEGATIVE_NUM(1)
        True
        >>> is_NON_NEGATIVE_NUM(0)
        True
        >>> is_NON_NEGATIVE_NUM(-1)
        False
        >>> is_NON_NEGATIVE_NUM(1.0)
        True
    """
    return is_NUM(x) and x >= 0


def is_NON_NEGATIVE_INT(x: Any) -> bool:
    """Check if x is a non-negative integer.

    Examples:
        >>> is_NON_NEGATIVE_INT(1)
        True
        >>> is_NON_NEGATIVE_INT(0)
        True
        >>> is_NON_NEGATIVE_INT(-1)
        False
        >>> is_NON_NEGATIVE_INT(1.0)
        False
    """
    return isinstance(x, int) and x >= 0


def is_PROPORTION(x: Any) -> bool:
    """Check if x is a proportion (between 0 and 1 inclusive).

    Examples:
        >>> is_PROPORTION(1)
        True
        >>> is_PROPORTION(0)
        True
        >>> is_PROPORTION(0.5)
        True
        >>> is_PROPORTION(-1)
        False
        >>> is_PROPORTION("foo")
        False
    """
    return is_NUM(x) and 0 <= x <= 1


def is_POSITIVE_TIMEDELTA(x: Any) -> bool:
    """Check if x is a positive timedelta.

    Examples:
        >>> is_POSITIVE_TIMEDELTA(np.timedelta64(1, "s"))
        True
        >>> is_POSITIVE_TIMEDELTA(np.timedelta64(0, "s"))
        False
        >>> is_POSITIVE_TIMEDELTA(np.timedelta64(1, "m"))
        True
        >>> is_POSITIVE_TIMEDELTA(np.timedelta64(-1, "s"))
        False
        >>> is_POSITIVE_TIMEDELTA(1)
        False
    """
    return bool(isinstance(x, np.timedelta64) and (x > 0))



import abc
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Generic, TypeVar

import numpy as np
import pytimeparse

from .types import (
    NON_NEGATIVE_NUM,
    POSITIVE_INT,
    POSITIVE_NUM,
    POSITIVE_TIMEDELTA,
    PROPORTION,
    is_NON_NEGATIVE_NUM,
    is_POSITIVE_INT,
    is_POSITIVE_NUM,
    is_POSITIVE_TIMEDELTA,
    is_PROPORTION,
)

T = TypeVar("T")


class Stringified(Generic[T], abc.ABC):
    @classmethod
    def _to_str(cls, x: Any) -> str:
        return str(x)

    @classmethod
    @abc.abstractmethod
    def _from_str(cls, x: T) -> Any:  # pragma: no cover
        raise NotImplementedError

    @property
    def _X(self) -> np.ndarray:
        return np.array([self._from_str(x) for x in self.X])

    @abc.abstractmethod
    def _validate(self):  # pragma: no cover
        raise NotImplementedError

    def __post_init__(self):
        if not all(isinstance(x, str) for x in self.X):
            self.X = [self._to_str(x) for x in self.X]

        self._validate(self._X)

        super().__post_init__()


@dataclass
class DiscreteGenerator:
    """A class to generate random numbers from a list of options with given frequencies.

    This is largely just for type safety and to ease specification of the various things that need to be
    sampled to generate a dataset.

    Attributes:
        X: The list of options to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: If the frequencies are not all positive, the lengths of X and freq are not equal, or no
            options are provided.

    Examples:
        >>> x = DiscreteGenerator([1, 2, 3], [1, 2, 3])
        >>> rng = np.random.default_rng(1)
        >>> x.rvs(10, rng)
        array([3, 3, 1, 3, 2, 2, 3, 2, 3, 1])
        >>> rng = np.random.default_rng(1)
        >>> x.rvs(10, rng)
        array([3, 3, 1, 3, 2, 2, 3, 2, 3, 1])
        >>> x.rvs(10, rng)
        array([3, 3, 2, 3, 2, 2, 1, 2, 2, 2])
        >>> rng = np.random.default_rng(1)
        >>> DiscreteGenerator([1, 2, 3]).rvs(10, rng)
        array([2, 3, 1, 3, 1, 2, 3, 2, 2, 1])
        >>> rng = np.random.default_rng(1)
        >>> DiscreteGenerator(['a', 'b', 'c']).rvs(10, rng)
        array(['b', 'c', 'a', 'c', 'a', 'b', 'c', 'b', 'b', 'a'], dtype='<U1')
        >>> DiscreteGenerator([1, 2], [-1, 1])
        Traceback (most recent call last):
            ...
        ValueError: All frequencies should be positive.
        >>> DiscreteGenerator([1, 2], [1, 2, 3])
        Traceback (most recent call last):
            ...
        ValueError: Equal numbers of frequencies and options must be provided. Got 3 and 2.
        >>> DiscreteGenerator([])
        Traceback (most recent call last):
            ...
        ValueError: At least one option should be provided. Got length 0.
    """

    X: list[Any]
    freq: list[NON_NEGATIVE_NUM] | None = None

    def __post_init__(self):
        if self.freq is None:
            self.freq = [1] * len(self.X)
        if not all(is_NON_NEGATIVE_NUM(f) for f in self.freq):
            raise ValueError("All frequencies should be positive.")
        if len(self.freq) != len(self.X):
            raise ValueError(
                "Equal numbers of frequencies and options must be provided. "
                f"Got {len(self.freq)} and {len(self.X)}."
            )
        if len(self.freq) == 0:
            raise ValueError("At least one option should be provided. Got length 0.")

    @property
    def _X(self) -> np.ndarray:
        return np.array(self.X)

    @property
    def p(self) -> np.ndarray:
        return np.array(self.freq) / sum(self.freq)

    def rvs(self, size: int, rng: np.random.Generator) -> np.ndarray:
        return rng.choice(self._X, size=size, p=self.p, replace=True)


class DatetimeGenerator(Stringified[np.datetime64], DiscreteGenerator):
    """A class to generate random datetimes.

    This merely applies type-checking to the DiscreteGenerator class.

    Attributes:
        X: The list of datetimes to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: In addition to the base class errors, if any of the options are not datetimes.

    Examples:
        >>> rng = np.random.default_rng(1)
        >>> DatetimeGenerator([np.datetime64("2021-01-01"), np.datetime64("2022-02-02")]).rvs(10, rng)
        array(['2022-02-02', '2022-02-02', '2021-01-01', '2022-02-02',
               '2021-01-01', '2021-01-01', '2022-02-02', '2021-01-01',
               '2022-02-02', '2021-01-01'], dtype='datetime64[D]')
        >>> DatetimeGenerator([1, 2])
        Traceback (most recent call last):
            ...
        ValueError: Could not convert '1' to a datetime.
        >>> DatetimeGenerator(["a"])
        Traceback (most recent call last):
            ...
        ValueError: Could not convert 'a' to a datetime.
    """

    X: list[str | np.datetime64]

    @classmethod
    def _from_str(cls, x: str) -> np.datetime64:
        """Convert a string to a datetime.

        Examples:
            >>> DatetimeGenerator._from_str("2021-01-01")
            np.datetime64('2021-01-01')
            >>> DatetimeGenerator._from_str("2021-01-01 00:00:00")
            np.datetime64('2021-01-01T00:00:00')
            >>> DatetimeGenerator._from_str("2021-01-01T00:00:00")
            np.datetime64('2021-01-01T00:00:00')
            >>> DatetimeGenerator._from_str("2021-01-01T00:00:00.000")
            np.datetime64('2021-01-01T00:00:00.000')
            >>> DatetimeGenerator._from_str("1")
            Traceback (most recent call last):
                ...
            ValueError: Could not convert '1' to a datetime.
        """
        formats = ["%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S.%f"]

        try:
            for fmt in formats:
                try:
                    _ = np.datetime64(datetime.strptime(x, fmt))
                    return np.datetime64(x)
                except Exception:
                    pass
            raise ValueError(f"Could not convert '{x}' to a datetime in allowed formats: {formats}.")
        except Exception as e:
            raise ValueError(f"Could not convert '{x}' to a datetime.") from e

    def _validate(self, X: list[Any]):
        pass


class PositiveTimeDeltaGenerator(Stringified[np.timedelta64], DiscreteGenerator):
    """A class to generate random positive timedeltas.

    This merely applies type-checking to the DiscreteGenerator class.

    Attributes:
        X: The list of timedeltas to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: In addition to the base class errors, if any of the options are not positive timedeltas.

    Examples:
        >>> rng = np.random.default_rng(1)
        >>> PositiveTimeDeltaGenerator([np.timedelta64(1, "s"), np.timedelta64(2, "s")]).rvs(10, rng)
        array([2, 2, 1, 2, 1, 1, 2, 1, 2, 1], dtype='timedelta64[s]')
        >>> PositiveTimeDeltaGenerator([1, 2])
        Traceback (most recent call last):
            ...
        ValueError: Could not convert 1 to a proper string.
        >>> PositiveTimeDeltaGenerator([np.timedelta64(1, "s"), np.timedelta64(-1, "s")])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be positive timedeltas. Got [np.timedelta64(-1,'s')].
        >>> PositiveTimeDeltaGenerator(["a"])
        Traceback (most recent call last):
            ...
        ValueError: Could not convert 'a' to a timedelta.
    """

    X: list[POSITIVE_TIMEDELTA | str]

    def _validate(self, X: list[Any]):
        if not all(is_POSITIVE_TIMEDELTA(x) for x in X):
            fails = [x for x in X if not is_POSITIVE_TIMEDELTA(x)]
            raise ValueError(f"All elements should be positive timedeltas. Got {fails}.")

    @classmethod
    def _to_str(cls, x: np.timedelta64) -> str:
        try:
            as_sec = x.astype("timedelta64[s]") / np.timedelta64(1, "s")
            return f"{as_sec}s"
        except Exception as e:
            raise ValueError(f"Could not convert {x} to a proper string.") from e

    @classmethod
    def _from_str(cls, x: str) -> np.timedelta64:
        try:
            return np.timedelta64(int(pytimeparse.parse(x) * 1e9), "ns").astype("timedelta64[s]")
        except Exception as e:
            raise ValueError(f"Could not convert '{x}' to a timedelta.") from e


class ProportionGenerator(DiscreteGenerator):
    """A class to generate random proportions.

    This merely applies type-checking to the DiscreteGenerator class.

    Attributes:
        X: The list of proportions to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: In addition to the base class errors, if any of the proportions are not numbers between 0
            and 1.

    Examples:
        >>> rng = np.random.default_rng(1)
        >>> ProportionGenerator([0, 1, 0.3]).rvs(10, rng)
        array([1. , 0.3, 0. , 0.3, 0. , 1. , 0.3, 1. , 1. , 0. ])
        >>> ProportionGenerator([1, 2])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be proportions (numbers between 0 and 1 inclusive).
        >>> ProportionGenerator(["a"])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be proportions (numbers between 0 and 1 inclusive).
    """

    X: list[PROPORTION]

    def __post_init__(self):
        if not all(is_PROPORTION(x) for x in self.X):
            raise ValueError("All elements should be proportions (numbers between 0 and 1 inclusive).")
        super().__post_init__()


class PositiveNumGenerator(DiscreteGenerator):
    """A class to generate random positive numbers.

    This merely applies type-checking to the DiscreteGenerator class.

    Attributes:
        X: The list of positive numbers to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: In addition to the base class errors, if any of the options are not positive numbers.

    Examples:
        >>> rng = np.random.default_rng(1)
        >>> PositiveNumGenerator([1, 2, 3.0]).rvs(10, rng)
        array([2., 3., 1., 3., 1., 2., 3., 2., 2., 1.])
        >>> PositiveNumGenerator([1, -1, 2])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be positive numbers.
        >>> PositiveNumGenerator([0])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be positive numbers.
        >>> PositiveNumGenerator(["a"])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be positive numbers.
    """

    X: list[POSITIVE_NUM]

    def __post_init__(self):
        if not all(is_POSITIVE_NUM(x) for x in self.X):
            raise ValueError("All elements should be positive numbers.")
        super().__post_init__()


class PositiveIntGenerator(PositiveNumGenerator):
    """A class to generate random positive integers.

    This merely applies type-checking to the DiscreteGenerator class.

    Attributes:
        X: The list of positive integers to sample from.
        freq: The frequency of each option. If None, all options are equally weighted.

    Raises:
        ValueError: In addition to the base class errors, if any of the options are not positive integers.

    Examples:
        >>> rng = np.random.default_rng(1)
        >>> PositiveIntGenerator([1, 2, 3]).rvs(10, rng)
        array([2, 3, 1, 3, 1, 2, 3, 2, 2, 1])
        >>> PositiveIntGenerator([0.1])
        Traceback (most recent call last):
            ...
        ValueError: All elements should be integers.
    """

    X: list[POSITIVE_INT]

    def __post_init__(self):
        if not all(is_POSITIVE_INT(x) for x in self.X):
            raise ValueError("All elements should be integers.")
        super().__post_init__()
